//
// XMM Science Analysis System
//
// Giuseppe Vacanti (European Space Agency) 
// April 20, 1999
//


#include <unistd.h> // C

#include <sstream> // C++
#include <map>
#include <cstdlib>
#include <cstdio>

#include <qlayout.h>
#include <qlabel.h>
#include <qwidget.h>
#include <qmessagebox.h>
#include <qmenubar.h>
#include <qpopupmenu.h>
#include <qfiledialog.h>
#include <qbutton.h>
#include <qlineedit.h>
#include <qstringlist.h> 

#include "CalviewGui.h"

#include "InstrumentBgroup.h"
#include "CcdBgroup.h"
#include "NodeBgroup.h"
#include "ModeBgroup.h"
#include "FilterBgroup.h"
#include "CcdTemperatureSelector.h"
#include "CameraTemperatureSelector.h"
#include "OcbSelector.h"
#include "EnergySelector.h"
#include "ThetaSelector.h"
#include "PhiSelector.h"
#include "ObservationStartTimeSelector.h"
#include "AccuracySelector.h"
#include "RandomizationSelector.h"
#include "SpectralOrderSelector.h"

#include "CcfLogger.h"

#include <errstr.h> // SAS
#include <ExceptionErrorHandler.h>
#include <CalView.h>
#include <CalServer.h>
#include <Tracer.h>

#include <Parameter.h>
#include <imgdisplay.h>

#include <qdialog.h>
#include <qtextview.h>
#include <qlayout.h>
#include <qpushbutton.h>

class Bla : public QDialog
{
public:
  Bla (QWidget *parent,
       const char *name
       ): QDialog(parent, name, true)
    {
      setCaption("Current Calibration File Constituents");
      _vbox = new QVBoxLayout(this, 20);
      _view = new QTextView(this);
      _view->setMinimumWidth(300);
      _view->setMinimumHeight(200);
      _ok = new QPushButton("Ok", this);
      _ok->setFixedSize(50,50);
      _vbox->addWidget(_view);
      _vbox->addWidget(_ok);
      connect(_ok, SIGNAL(released()), this, SLOT(accept()));

    }
  void setText(const QString& txt ) { _view->setText(txt);}
private:
  QTextView* _view;
  QPushButton *_ok;
  QBoxLayout *_vbox;
  QLabel* _label;
};


// calview_set is where we store the intermediate output before
// passing it on to the viewer. It is a global so that it can be
// unlinked by calview_cleanup, in case calview is terminated.

//static string calview_set = "calview.out";

static vector<string> calview_sets;

static void calview_cleanup()
{
  PROC
    for(unsigned k=0; k<calview_sets.size(); ++k)
      unlink(calview_sets[k].c_str());
  ENDPROC
    //  PROC unlink(calview_set.c_str()); ENDPROC
}
  
CalviewGui::CalviewGui(const char *name)
  : SasTaskWindow(name),
    _ds_counter(0)
{
  PROC

  atexit(calview_cleanup);

  QWidget* fooBar = new QWidget(this);
  QGridLayout * vbox = new QGridLayout(fooBar,20,2);
  setCentralWidget(fooBar);

  // Cal state

  int row = 0;

  vbox->addMultiCellWidget(new QLabel("Calibration State Editor",fooBar),
			   row,row,0,1,AlignCenter);

  InstrumentBgroup * ibg = new InstrumentBgroup(fooBar);
  QLabel * ibgl = new QLabel("Instrument",fooBar);
  ibgl->setFixedSize(ibgl->sizeHint());
  vbox->addWidget(ibgl,++row,0);
  vbox->addWidget(ibg,row,1);

  CcdBgroup * cbg = new CcdBgroup(fooBar);
  QLabel * cbgl = new QLabel("CCD",fooBar);
  cbgl->setFixedSize(cbgl->sizeHint());
  vbox->addWidget(cbgl,++row,0);
  vbox->addWidget(cbg,row,1);

  NodeBgroup * nbg = new NodeBgroup(fooBar);
  QLabel * nbgl = new QLabel("Node",fooBar);
  nbgl->setFixedSize(nbgl->sizeHint());
  vbox->addWidget(nbgl,++row,0);
  vbox->addWidget(nbg,row,1);

  FilterBgroup * fbg = new FilterBgroup(fooBar);
  QLabel * fbgl = new QLabel("Filter",fooBar);
  fbgl->setFixedSize(fbgl->sizeHint());
  vbox->addWidget(fbgl,++row,0);
  vbox->addWidget(fbg,row,1);

  ModeBgroup * mbg = new ModeBgroup(fooBar);
  QLabel * mbgl = new QLabel("Mode",fooBar);
  mbgl->setFixedSize(mbgl->sizeHint());
  vbox->addWidget(mbgl,++row,0);
  vbox->addWidget(mbg,row,1);

  CcdTemperatureSelector * cts = new CcdTemperatureSelector(fooBar);
  QLabel * ctsl = new QLabel("CCD Temperature (K)", fooBar);
  ctsl->setFixedSize(ctsl->sizeHint());
  vbox->addWidget(ctsl,++row,0);
  vbox->addWidget(cts,row,1);

  QLabel * catsl = new QLabel("Camera Temperature (K)",fooBar);
  CameraTemperatureSelector * cats = new CameraTemperatureSelector(fooBar);
  catsl->setFixedSize(catsl->sizeHint());
  vbox->addWidget(catsl,++row,0);
  vbox->addWidget(cats,row,1);

  QLabel * ocbsl = new QLabel("On-Chip Binning",fooBar);
  OcbSelector * ocbs = new OcbSelector(fooBar);
  ocbsl->setFixedSize(ocbsl->sizeHint());
  vbox->addWidget(ocbsl,++row,0);
  vbox->addWidget(ocbs,row,1);

  QLabel * ostsl = new QLabel("Date",fooBar);
  ObservationStartTimeSelector * osts = new ObservationStartTimeSelector(fooBar);
  ostsl->setFixedSize(ostsl->sizeHint());
  vbox->addWidget(ostsl,++row,0);
  vbox->addWidget(osts,row,1);

  QLabel * asl = new QLabel("Accuracy Level",fooBar);
  AccuracySelector * as = new AccuracySelector(fooBar);
  asl->setFixedSize(asl->sizeHint());
  vbox->addWidget(asl,++row,0);
  vbox->addWidget(as,row,1);

  QLabel * rsl = new QLabel("Randomization",fooBar);
  RandomizationSelector * rs = new RandomizationSelector(fooBar);
  rsl->setFixedSize(rsl->sizeHint());
  vbox->addWidget(rsl,++row,0);
  vbox->addWidget(rs,row,1);

  // CalView state
  row++;
  vbox->addMultiCellWidget(new QLabel("Calibration Viewer State Editor",fooBar),
			   row,row,0,1,AlignCenter);
  
  EnergySelector * es = new EnergySelector(fooBar);
  QLabel * esl = new QLabel("Energy (eV)", fooBar);
  esl->setFixedSize(esl->sizeHint());
  vbox->addWidget(esl,++row,0);
  vbox->addWidget(es,row,1);

  ThetaSelector * ts = new ThetaSelector(fooBar);
  QLabel * tsl = new QLabel("Theta (arcsec)", fooBar);
  tsl->setFixedSize(tsl->sizeHint());
  vbox->addWidget(tsl,++row,0);
  vbox->addWidget(ts,row,1);

  PhiSelector * ps = new PhiSelector(fooBar);
  QLabel * psl = new QLabel("Phi (deg)", fooBar);
  psl->setFixedSize(psl->sizeHint());
  vbox->addWidget(psl,++row,0);
  vbox->addWidget(ps,row,1);

  SpectralOrderSelector *sos = new SpectralOrderSelector(fooBar);
  QLabel *sosl = new QLabel("Order", fooBar);
  sosl->setFixedSize(sosl->sizeHint());
  vbox->addWidget(sosl,++row,0);
  vbox->addWidget(sos,row,1);

// This is the size of the largest widget we have in column 1
//    QWidget * sizeref = mbg;
//    ibg->setFixedSize(sizeref->sizeHint());
//    cbg->setFixedSize(sizeref->sizeHint());
//    nbg->setFixedSize(sizeref->sizeHint());
//    fbg->setFixedSize(sizeref->sizeHint());
//    mbg->setFixedSize(sizeref->sizeHint());
//    cts->setFixedSize(sizeref->sizeHint());
//    cats->setFixedSize(sizeref->sizeHint());
//    ocbs->setFixedSize(sizeref->sizeHint());
//    osts->setFixedSize(sizeref->sizeHint());
//    as->setFixedSize(sizeref->sizeHint());
//    rs->setFixedSize(sizeref->sizeHint());

//    es->setFixedSize(sizeref->sizeHint());

  // CCF logger
  ++row;
  vbox->addMultiCellWidget(new QLabel("CCF Access Log", fooBar),
			   row, row, 0, 1, AlignCenter);

  ++row;
  CcfLogger *cl = new CcfLogger(fooBar);
  vbox->addMultiCellWidget(cl, row, row, 0, 1, 0);

  // A new viewable triggers the update of the CCF access log window
  connect(this, SIGNAL(viewerComplete()), cl, SLOT(update()));
  
  // An instrument update touches all the other CAL state variables ...
  connect(ibg, SIGNAL(calStateIsUpdated()), cbg, SLOT(update()));
  connect(ibg, SIGNAL(calStateIsUpdated()), nbg, SLOT(update()));
  connect(ibg, SIGNAL(calStateIsUpdated()), fbg, SLOT(update()));
  connect(ibg, SIGNAL(calStateIsUpdated()), mbg, SLOT(update()));
  connect(ibg, SIGNAL(calStateIsUpdated()), cts, SLOT(update()));
  connect(ibg, SIGNAL(calStateIsUpdated()), cats, SLOT(update()));
  connect(ibg, SIGNAL(calStateIsUpdated()), ocbs, SLOT(update()));
  connect(ibg, SIGNAL(calStateIsUpdated()), osts, SLOT(update()));
  connect(ibg, SIGNAL(calStateIsUpdated()), as, SLOT(update()));
  connect(ibg, SIGNAL(calStateIsUpdated()), rs, SLOT(update()));

  // ... and brings in a new list of viewables
  connect(ibg, SIGNAL(calStateIsUpdated()), this, SLOT(updateViewables()));

  // A CCD update touches node, chip temperature, and on-chip binnig factor
  connect(cbg, SIGNAL(calStateIsUpdated()), nbg, SLOT(update()));
  connect(cbg, SIGNAL(calStateIsUpdated()), cts, SLOT(update()));
  connect(cbg, SIGNAL(calStateIsUpdated()), ocbs, SLOT(update()));

  // Any CAL state update causes the display of the new CAL state
  connect(ibg, SIGNAL(calStateIsUpdated()), this, SLOT(showState()));
  connect(cbg, SIGNAL(calStateIsUpdated()), this, SLOT(showState()));
  connect(nbg, SIGNAL(calStateIsUpdated()), this, SLOT(showState()));
  connect(fbg, SIGNAL(calStateIsUpdated()), this, SLOT(showState()));
  connect(mbg, SIGNAL(calStateIsUpdated()), this, SLOT(showState()));
  connect(cts, SIGNAL(calStateIsUpdated()), this, SLOT(showState()));
  connect(cats, SIGNAL(calStateIsUpdated()), this, SLOT(showState()));
  connect(ocbs, SIGNAL(calStateIsUpdated()), this, SLOT(showState()));
  connect(osts, SIGNAL(calStateIsUpdated()), this, SLOT(showState()));
  connect(as, SIGNAL(calStateIsUpdated()), this, SLOT(showState()));
  connect(rs, SIGNAL(calStateIsUpdated()), this, SLOT(showState()));

  menuBarInit();
  this->resize(this->sizeHint());

  ENDPROC
}

QMenuBar * CalviewGui::menuBarInit()
{
  PROC
  QPopupMenu *file = new QPopupMenu;
  file->insertItem("E&xit", this, SLOT(quit()), CTRL+Key_X);

  _view = new QPopupMenu;  
  updateViewables();
  connect(_view, SIGNAL(activated(int)), this, SLOT(viewDispatcher(int)));

  QPopupMenu * ccf = new QPopupMenu;
  ccf->insertItem("Set a new CCF directory", this, 
		  SLOT(setNewCcfDirectory()));
  ccf->insertItem("Set a new CCF calibration index file", this,
		  SLOT(setNewCcfCif()));
  ccf->insertItem("Add a local CCF component", this,
		  SLOT(insertCcf()));
  ccf->insertItem("List CCF components in use", this,
  		  SLOT(listCcf()));

  menuBar()->insertItem("&File", file,-1,0);
  menuBar()->insertItem("&CCF", ccf,-1,1);
  menuBar()->insertItem("&View", _view,-1,2);

  RETURN menuBar();
}


void CalviewGui::viewDispatcher(int i)
{
  PROC
//    cerr << "viewDispatcher called with " << i << endl;
//    cerr << _call_back[i]->name() << endl;
//    pid_t cvid = getpid();
//    ostrstream os;
//    os << "calview_" << cvid ;
//    string set = os.str();
 
  // Flush all CCF components to disk to force the server to access
  // them a gain and update the CcfAccessLog. 
  calServer.atomServer()->restart();
  CcfAccessLog &ccflog = calServer.ccfAccessLog();
  // Empty the CcfAccessLog, so we can monitor the Ccf accesses for
  // the next viewable.
  ccflog.reset();

  //  string set(tmpnam(0));

  string calview_set = calview_ds();

  ExceptionErrorHandler local;
  ErrorHandler* oldHandler = errHandler.handler(&local);
  CalViewProductDescriptor p("dummy");
  try {
    p = _call_back[i]->view(calview_set);

    string command;
    switch (p.type()) {
    case CalViewProductDescriptor::XMGR_PARM:
      command = "(xmgrace " + calview_set + " ; rm -f " + calview_set + ")&";
      system(command.c_str());
      break;

    case CalViewProductDescriptor::FITS_IMG:
      openParameters("imgdisplay");
      setParameter("withimagefile", true);
      setParameter("imagefile", calview_set);
      task(imgdisplay);
      sleep(1); // needed, as there is a small delay starting ds9
      //    command = "ds9 " + calview_set;
      //    system(command.c_str());
      break;
      
    case CalViewProductDescriptor::ASCII:
      errstr << "Viewing of ASCII data not supported yet. Output in "
	+ calview_set << Sas::message(Msg::App, Msg::Sparse);
      break;
      
    default:
      errstr << "Should not be here. Report this."
	     << Sas::error("UnknonwType::viewDispatcher");
    }
    emit viewerComplete();    
  }
  catch (SasException::Error e) {
//      errstr << e.code << " == " << e.message
//  	   << Sas::warning("CalView::CalError");
    QString text("");
    text.append("An error was detected while trying to view ");
    text.append("\n");
    text.append(_call_back[i]->name().c_str());
    text.append("\n");
    text.append("The error code was: ");
    text.append(e.code.c_str());
    text.append("\n");
    text.append("The error message was: ");
    text.append(e.message.c_str());
    text.append("\n");
    QMessageBox::information(this, "calview", text);
  }
  catch (SasException::Fatal e) {
    QString text("");
    text.append("An error was detected while trying to view ");
    text.append("\n");
    text.append(_call_back[i]->name().c_str());
    text.append("\n");
    text.append("The error code was: ");
    text.append(e.code.c_str());
    text.append("\n");
    text.append("The error message was: ");
    text.append(e.message.c_str());
    text.append("\n");
    QMessageBox::information(this, "calview", text);
  }
  
  errHandler.handler(oldHandler);

  ENDPROC
}  
    
CalviewGui::~CalviewGui()
{
  PROC
  ENDPROC
}


void CalviewGui::setNewCcfDirectory()
{
  PROC
    //  QString d = QFileDialog::getExistingDirectory();
    QFileDialog * fd = new QFileDialog(this,0,TRUE);
  fd->setMode(QFileDialog::Directory);
  fd->setCaption("Select a CCF directory ...");
  if(fd->exec() == QDialog::Accepted){
    QString d = fd->selectedFile();
    if(!d.isNull())
      calServer.setCcf(string(d.ascii()));
  }
  ENDPROC
}

void CalviewGui::setNewCcfCif()
{
  PROC
  QFileDialog * fd = new QFileDialog(this,0,TRUE);
  QStringList s;
  s << "*.cif" << "*.CIF" << "*" ;
  fd->setFilters(s);
  fd->setMode(QFileDialog::ExistingFile);
  fd->setCaption("Select a CCF calibration index file ...");
  if(fd->exec() == QDialog::Accepted){
    QString d = fd->selectedFile();
  //QString d = QFileDialog::getOpenFileName(".","*.cif",this);
    if(!d.isNull())
      calServer.setCcf(string(d.ascii()));
  }
  ENDPROC
}

void CalviewGui::insertCcf()
{
  PROC
    //	  QString d = QFileDialog::getOpenFileName(".","*.CCF",this);
  QFileDialog * fd = new QFileDialog(this,0,TRUE);
  QStringList s;
  s << "*.ccf" << "*.CCF" << "*" ;
  fd->setFilters(s);
  fd->setMode(QFileDialog::ExistingFile);
  fd->setCaption("Select one CCF component ...");
  if(fd->exec() == QDialog::Accepted){
    QString d = fd->selectedFile();
    if(!d.isNull())
      calServer.ccf()->registerNewCifEntry(string(d.ascii())); 
  }
  ENDPROC
}

void CalviewGui::listCcf()
{
  PROC
  typedef pair<Xmm::Instrument, Ccf::CalDataTypes> CcfConstituent;
  typedef map<CcfConstituent, CifEntry *, less<CcfConstituent> > Cif;

  const Cif& cif = calServer.ccf()->cif();
  QString text("");
  for(Cif::const_iterator k=cif.begin(); k!=cif.end(); ++k){
    text.append("\n");
    text.append((*k).second->fileName().c_str());
  }
  Bla qtw(this, "Current Calibration File Constituents");
  qtw.setText(text);
  qtw.exec();
  ENDPROC
}

void CalviewGui::notImplemented()
{
  PROC
  ENDPROC
}

void CalviewGui::donothing()
{
  PROC
  ENDPROC
}

void CalviewGui::processViewable(const CalViewable* v,
				 QPopupMenu* qpm
				 )
{
  PROC
  int i;
  if(v->subViewables() == 0){
    // There are no (further) subviewables, register and return
    i = qpm->insertItem(v->name().c_str(), this, SLOT(donothing()));
    errstr << "Registering viewable " << v->name() << " in slot "
	   << i << Sas::message(Msg::App, Msg::Noisy);
    _call_back[i] = v;
  } else {
    // There are subviewables
    QPopupMenu* q = new QPopupMenu(this);
    connect(q, SIGNAL(activated(int)), this, SLOT(viewDispatcher(int)));
    i = qpm->insertItem(v->name().c_str(), q);
    errstr << "Registering viewable " << v->name() << " in slot "
	   << i << Sas::message(Msg::App, Msg::Noisy);
    for(VI sv = v->subViewable().begin();
	sv != v->subViewable().end(); sv++){
      processViewable(*sv, q);
    }
  }
  ENDPROC
}

void CalviewGui::updateViewables()
{
  PROC
  _view->clear();
  _call_back.clear();
  vector<const CalViewable *> viewables = calViewServer.viewable();
  for ( vector<const CalViewable *>::iterator v = viewables.begin(); v != viewables.end(); v++) {
    processViewable(*v, _view);
  }
  ENDPROC
}

void CalviewGui::showState()
{
  PROC
  SAS_DEBUG(
  cerr << "-- CAL state --" << endl;
  cerr << calServer.getState() << endl;
  )
  ENDPROC
}

string CalviewGui::calview_ds()
{
  PROC
  stringstream s;
  s << "calview_" << _ds_counter++ << "_" << getpid() << ".out" ;
  string name = s.str();
  calview_sets.push_back(name);
  RETURN name;
}
  
